/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 * 
 */

package ant.compiler;

import ant.compiler.parser.*;
import ant.compiler.ast.Tree;
import ant.compiler.ast.TreePrinter;
import ant.compiler.trafo.*;
import ant.compiler.asm.*;
import ant.compiler.util.ErrorLog;
import ant.compiler.gen.*;

import java.io.*;
import java.util.ArrayList;

/**
 * AntCompiler main class.
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class AntCompiler {
    public static final String EXTENSION_ANT = ".ant";
    public static final String EXTENSION_ACP = ".acp";

    public static boolean debug = false;
    public static boolean pretty = false;

    public static void main(String[] args) throws Exception {
	for (int argc = 0; argc < args.length; argc++) {
	    String arg = args[argc];

	    if (arg.startsWith("-")) {
		if (arg.equals("-debug")) {
		    debug = true;
		} 

		else if (arg.equals("-pretty")) {
		    pretty = true;
		}

		else {
		    System.err.println("error: unknown option: '" + arg + "'");
		    System.exit(1);
		}
	    } else {
		compile(arg);
	    }
	}
    }

    public static boolean compile(String filename) throws IOException {
	FileInputStream in = new FileInputStream(filename);

	ErrorLog log = new ErrorLog(filename);

	AntParser p = new AntParser(in);

	Tree tree;
	try {
	    tree = p.stmt();
	} catch (ParseException ex) {
	    System.err.println(ex.getMessage());
	    return false;
	}

	if (pretty) {
	    TreePrinter.dumpTree(tree, /*debug*/ false);
	    return true;
	}

	SemanticAnalyzer semantic = new SemanticAnalyzer(log);

	tree.visit(semantic);
	if (debug) {
	    System.err.println("tree after initial semantic analysis");
	    TreePrinter.dumpTree(tree, debug);
	}
	if (log.hasErrors()) return false;

	ProcTransformer proctrafo = new ProcTransformer(log, tree);
	tree.visit(proctrafo);
	tree = proctrafo.getResult();
	if (log.hasErrors()) return false;

	tree.visit(semantic);
	if (debug) {
	    System.err.println("tree after procedure transformation");
	    TreePrinter.dumpTree(tree, debug);
	}
	if (log.hasErrors()) return false;
	
	AssignTransformer assigntrafo = new AssignTransformer();
	tree.visit(assigntrafo);
	tree = assigntrafo.getResult();

	tree.visit(semantic);
	if (debug) {
	    System.err.println("tree after assignment transformation");
	    TreePrinter.dumpTree(tree, debug);
	}
	if (log.hasErrors()) return false;

	VarTransformer vartrafo = new VarTransformer(log, tree);
	tree.visit(vartrafo);

	if (debug) {
	    System.err.println("tree after variable transformation");
	    TreePrinter.dumpTree(tree, debug);
	}
	if (log.hasErrors()) return false;

	AsmSymbols symbols = new AsmSymbols();
	CodeGenerator generator = new CodeGenerator(symbols);

	generator.reset();
	tree.visit(generator);
	Asm root = generator.getResult();

	symbols.resolveLabels();

	if (debug) {
	    System.err.println("code before nop resolution");
	    dumpAsm(root);
	}

	NopAnalyzer analyser = new NopAnalyzer(log);
	root.visit(analyser);

	NopEliminator eliminator = new NopEliminator(log);
	root.visit(eliminator);
	root = eliminator.getResult();

	AddressResolver address = new AddressResolver();
	root.visit(address);
	ArrayList commands = address.getCommands();

	if (! log.hasErrors()) {

	    StateMinimizer minimizer = new StateMinimizer(commands);
	    root = minimizer.minimize();

	    address.reset();
	    root.visit(address);
	    commands = address.getCommands();
	}

	PrintWriter out = new PrintWriter(
	    new FileWriter(
		(filename.endsWith(EXTENSION_ACP) ? 
		    filename.substring(0, filename.length() - EXTENSION_ACP.length()) : 
		    filename) + EXTENSION_ANT));
	
	AsmPrinter printer = new AsmPrinter(log, out, debug);
	for (int cnt = commands.size(), n = 0; n < cnt; n++) {
	    Asm op = (Asm) commands.get(n);
	    op.visit(printer);
	}

	out.flush();
	out.close();

	if (log.hasErrors()) return false;

	return true;
    }

    private static void dumpAsm(Asm root) {
	System.err.println("--- begin asm ---");
	PrintWriter out = new PrintWriter(System.err);
	AsmPrinter printer = new AsmPrinter(null, out, true);

	AddressResolver address = new AddressResolver();
	root.visit(address);
	ArrayList commands = address.getCommands();
	
	for (int cnt = commands.size(), n = 0; n < cnt; n++) {
	    Asm op = (Asm) commands.get(n);
	    op.visit(printer);
	}

	out.flush();
	System.err.println("--- end asm ---");
    }
}
