/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.asm;

import ant.compiler.ast.TurnDirection;
import ant.compiler.ast.SenseDirection;
import ant.compiler.ast.Condition;

import java.util.HashSet;


/**
 * Ant assembler with inner classes that represents the basic
 * statments for the ant simulator.
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public abstract class Asm {
    int addr;

    public abstract void setNext(Asm op);

    public abstract void visit(Visitor v);

    /**
     * Abstract super class for all sequential {@link Asm}
     * statements. Sequential statements have only a single successor.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static abstract class Sequential extends Asm {
	public Asm next;

	public Sequential() {
	    this.next = null;
	}

	public Sequential(Asm next) {
	    this.next = next;
	}

	public void setNext(Asm next) {
	    this.next = next;
	}
    }

    /**
     * Abstract super class for all conditional {@link Asm}
     * statements. Conditional statements have to possible successors.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static abstract class Conditional extends Asm {
	public Asm nextThen;
	public Asm nextElse;

	public Conditional(Asm nextThen, Asm nextElse) {
	    this.nextThen = nextThen;
	    this.nextElse = nextElse;
	}

	public void setNext(Asm next) {
	    this.nextThen = next;
	    this.nextElse = next;
	}
    }

    /**
     * {@link Asm} statement that does nothing. Such statements are
     * removed before writing the final code.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Nop extends Sequential {
	Asm replacement;

	public Nop() {
	}

	public Nop(Asm next) {
	    super(next);
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the special test
     * <code>HitWall</code>. This statement is joined with a
     * preceeding {@link Asm.Move} statement before writing the final
     * code.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class NopWallTest extends Conditional {
	public NopWallTest(Asm nextThen, Asm nextElse) {
	    super(nextThen, nextElse);
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the special test
     * <code>FoundFood</code>. This statement is joined with a
     * preceeding {@link Asm.PickUp} statement before writing the final
     * code.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class NopFoodTest extends Conditional {
	public NopFoodTest(Asm nextThen, Asm nextElse) {
	    super(nextThen, nextElse);
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>sense</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Sense extends Conditional {
	SenseDirection direction;
	Condition condition;

	public Sense(SenseDirection direction, Condition condition, Asm nextThen, Asm nextElse) {
	    super(nextThen, nextElse);
	    this.direction = direction;
	    this.condition = condition;
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof Sense)) return false;
	    Sense other = (Sense) _other;

	    if (this.direction != other.direction) return false;
	    if (this.condition != other.condition) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>mark</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Mark extends Sequential {
	int nr;

	public Mark(int nr, Asm next) {
	    super(next);

	    this.nr = nr;
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof Mark)) return false;
	    Mark other = (Mark) _other;

	    if (this.nr != other.nr) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>unmark</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Unmark extends Sequential {
	int nr;

	public Unmark(int nr, Asm next) {
	    super(next);

	    this.nr = nr;
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof Unmark)) return false;
	    Unmark other = (Unmark) _other;

	    if (this.nr != other.nr) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>pickup</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class PickUp extends Conditional {
	public PickUp(Asm next) {
	    super(next, next);
	}

	public PickUp(Asm nextThen, Asm nextElse) {
	    super(nextThen, nextElse);
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof PickUp)) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>drop</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Drop extends Sequential {
	public Drop(Asm next) {
	    super(next);
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof Drop)) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>turn</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Turn extends Sequential { 
	TurnDirection direction;

	public Turn(TurnDirection direction, Asm next) {
	    super(next);

	    this.direction = direction;
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof Turn)) return false;
	    Turn other = (Turn) _other;

	    if (this.direction != other.direction) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>move</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Move extends Conditional {
	public Move(Asm next) {
	    this(next, next);
	}

	public Move(Asm nextThen, Asm nextElse) {
	    super(nextThen, nextElse);
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof Move)) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * {@link Asm} statement representing the <code>flip</code>
     * instruction.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Flip extends Conditional {
	int probability;

	public Flip(int probability, Asm next) {
	    this(probability, next, next);
	}

	public Flip(int probability, Asm nextThen, Asm nextElse) {
	    super(nextThen, nextElse);
	    this.probability = probability;
	}

	public boolean equals(Object _other) {
	    if (! (_other instanceof Flip)) return false;
	    Flip other = (Flip) _other;

	    if (this.probability != other.probability) return false;

	    return true;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Visitor interface implemented by all {@link Asm} statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public interface Visitor {
	public void visit(Nop op);
	public void visit(NopWallTest op);
	public void visit(NopFoodTest op);
	public void visit(Sense op);
	public void visit(Mark op);
	public void visit(Unmark op);
	public void visit(PickUp op);
	public void visit(Drop op);
	public void visit(Turn op);
	public void visit(Move op);
	public void visit(Flip op);
    }

    /**
     * Default implementation of the {@link Asm.Visitor} interface.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static abstract class AbstractVisitor implements Visitor {
	public void visitAsm(Asm op) {}
	public void visitSequential(Sequential op) {visitAsm(op); }
	public void visitConditional(Conditional op) {visitAsm(op); }

	public void visit(Nop op) {visitSequential(op); }
	public void visit(Mark op) {visitSequential(op); }
	public void visit(Unmark op) {visitSequential(op); }
	public void visit(Drop op) {visitSequential(op); }
	public void visit(Turn op) {visitSequential(op); }

	public void visit(NopWallTest op) {visitConditional(op); }
	public void visit(NopFoodTest op) {visitConditional(op); }
	public void visit(Sense op) {visitConditional(op); }
	public void visit(PickUp op) {visitConditional(op); }
	public void visit(Move op) {visitConditional(op); }
	public void visit(Flip op) {visitConditional(op); }
    }

    /**
     * Abstract depth first search visitor implementation of the
     * {@link Asm} statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static abstract class AbstractDFSVisitor extends AbstractVisitor {
	HashSet ops = new HashSet();

	public void reset() {
	    ops.clear();
	}

	protected boolean visited(Asm op) {
	    return ops.contains(op);
	}

	protected void markVisited(Asm op) {
	    ops.add(op);
	}

	protected void descendSequential(Sequential op) {
	    if (! visited(op)) {
		markVisited(op);
		op.next.visit(this);
	    }
	}

	protected void descendConditional(Conditional op) {
	    if (! visited(op)) {
		markVisited(op);
		op.nextThen.visit(this);
		op.nextElse.visit(this);
	    }
	}
    }

    /**
     * Depth first search visitor that performs a postorder graph
     * traversal.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static abstract class DFSPostorderVisitor extends AbstractDFSVisitor {
	public void visitSequential(Sequential op) {
	    descendSequential(op);
	    super.visitSequential(op);
	}

	public void visitConditional(Conditional op) {
	    descendConditional(op);
	    super.visitConditional(op);
	}
    }

    /**
     * Depth first search visitor that performs a preorder graph
     * traversal.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static abstract class DFSPreorderVisitor extends AbstractDFSVisitor {
	public void visitSequential(Sequential op) {
	    super.visitSequential(op);
	    descendSequential(op);
	}

	public void visitConditional(Conditional op) {
	    super.visitConditional(op);
	    descendConditional(op);
	}
    }
}
