/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.asm;

import ant.compiler.asm.Asm.*;
import ant.compiler.util.ErrorLog;
import ant.compiler.util.Printer;

import java.io.PrintWriter;

/**
 * Visitor for assembler statements in the {@link Asm} class.
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class AsmPrinter extends Printer implements Asm.Visitor {
    ErrorLog log;
    PrintWriter out;

    final boolean debug;

    public AsmPrinter(ErrorLog log, PrintWriter out, boolean debug) {
	this.log = log;
	this.out = out;
	this.debug = debug;
    }

    public void visit(Nop op) {
	print("Nop " + toStringTarget(op, op.next));
	printAddrComment(op);
    }

    public void visit(NopWallTest op) {
	print("NopWallTest " + 
	    toStringTarget(op, op.nextThen) + " " + 
	    toStringTarget(op, op.nextElse));
	printAddrComment(op);
    }

    public void visit(NopFoodTest op) {
	print("NopFoodTest " + 
	    toStringTarget(op, op.nextThen) + " " + 
	    toStringTarget(op, op.nextElse));
	printAddrComment(op);
    }

    public void visit(Sense op)	{
	print(
	    "Sense " + getLabel(op.direction) + " " + 
	    toStringTarget(op, op.nextThen) + " " + 
	    toStringTarget(op, op.nextElse) + " " + 
	    getLabel(op.condition) + 
	    (hasMarker(op.condition) ? 
  	        " " + getMarker(op.condition) :
	        ""));
	printAddrComment(op);
    }

    public void visit(Mark op) {
	print("Mark " + op.nr + " " + toStringTarget(op, op.next));
	printAddrComment(op);
    }

    public void visit(Unmark op) {
	print("Unmark " + op.nr + " " + toStringTarget(op, op.next));
	printAddrComment(op);
    }

    public void visit(PickUp op) {
	print("PickUp " + 
	    toStringTarget(op, op.nextThen) + " " + 
	    toStringTarget(op, op.nextElse));
	printAddrComment(op);
    }

    public void visit(Drop op) {
	print("Drop " + toStringTarget(op, op.next));
	printAddrComment(op);
    }

    public void visit(Turn op) {
	print("Turn " + 
	    getLabel(op.direction) + " " + 
	    toStringTarget(op, op.next));
	printAddrComment(op);
    }

    public void visit(Move op) {
	print("Move " + 
	    toStringTarget(op, op.nextThen) + " " + 
	    toStringTarget(op, op.nextElse));
	printAddrComment(op);
    }

    public void visit(Flip op) {
	print("Flip " + op.probability + " " + 
	    toStringTarget(op, op.nextThen) + " " + 
	    toStringTarget(op, op.nextElse));
	printAddrComment(op);
    }

    private void printAddrComment(Asm op) {
	if (debug) {
	    tab(25);
	    println("\t ; " + op.addr);
	} else {
	    println();
	}
    }

    int column = 0;

    private void print(String s) {
	column += s.length();
	out.print(s);
    }

    private void println() {
	column = 0;
	out.println();
    }

    private void println(String s) {
	column = 0;
	out.println(s);
    }

    private void tab(int column) {
	while (this.column < column) {
	    print(" ");
	}
    }

    private String toStringTarget(Asm source, Asm target) {
	if (target == null) {
	    if (log != null) 
		log.error("missing target for operation " + source.addr);
	    return "<null>";
	} else {
	    return Integer.toString(target.addr);
	}
    }
}
