/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.ast;

import ant.compiler.ast.Tree.*;


/**
 * Visitor that copies a tree fragment. 
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class CopyVisitor extends AbstractVisitor {
    private Tree copy;

    protected void setCopy(Tree copy) {
	this.copy = copy;
    }

    public Tree getCopy() {
	return this.copy;
    }

    public void visit(Label tree) {
	tree.stmt.visit(this);
	Tree stmt = getCopy();

	setCopy(new Label(tree.name, stmt));
    }

    public void visit(Block tree) {
	Tree[] stmts = new Tree[tree.stmts.length];
	for (int n = 0; n < tree.stmts.length; n++) {
	    tree.stmts[n].visit(this);
	    stmts[n] = getCopy();
	}

	setCopy(new Block(stmts));
    }

    public void visit(If tree) {
	tree.expr.visit(this);
	Expression expression = (Expression) getCopy();

	tree.thenStmt.visit(this);
	Tree thenStmt = getCopy();

	tree.elseStmt.visit(this);
	Tree elseStmt = getCopy();

	setCopy(new If(expression, thenStmt, elseStmt));
    }

    public void visit(Goto tree) {
	setCopy(new Goto(tree.name));
    }

    public void visit(Proc tree) {
	tree.body.visit(this);
	Block body = (Block) getCopy();

	Var[] params = new Var[tree.params.length];
	for (int n = 0; n < params.length; n++) {
	    tree.params[n].visit(this);
	    params[n] = (Var) getCopy();
	}

	setCopy(new Proc(tree.name, params, body));
    }

    public void visit(Call tree) {
	Expression[] args = new Expression[tree.args.length];
	for (int n = 0; n < tree.args.length; n++) {
	    tree.args[n].visit(this);
	    args[n] = (Expression) getCopy();
	}

	setCopy(new Call(tree.name, args));
    }

    public void visit(MarkCommand tree) {
	setCopy(new MarkCommand(tree.marker));
    }

    public void visit(UnmarkCommand tree) {
	setCopy(new UnmarkCommand(tree.marker));
    }

    public void visit(TurnCommand tree) {
	setCopy(new TurnCommand(tree.direction));
    }

    public void visit(MoveCommand tree) {
	setCopy(new MoveCommand());
    }

    public void visit(PickUpCommand tree) {
	setCopy(new PickUpCommand());
    }

    public void visit(DropCommand tree) {
	setCopy(new DropCommand());
    }

    public void visit(Var tree) {
	setCopy(new Var(tree.name));
    }

    public void visit(Set tree) {
	setCopy(new Set(tree.name, tree.value));
    }

    public void visit(Assign tree) {
	tree.expr.visit(this);
	Expression expr = (Expression) getCopy();

	setCopy(new Assign(tree.name, expr));
    }

    public void visit(Test tree) {
	setCopy(new Test(tree.name));
    }

    public void visit(Const tree) {
	setCopy(new Const(tree.value));
    }

    public void visit(NotExpression tree) {
	tree.sub.visit(this);
	Expression sub = (Expression) getCopy();

	setCopy(new NotExpression(sub));
    }

    public void visit(ConditionTest tree) {
	setCopy(new ConditionTest(tree.direction, tree.condition));
    }

    public void visit(WallTest tree) {
	setCopy(new WallTest());
    }

    public void visit(FoodTest tree) {
	setCopy(new FoodTest());
    }

    public void visit(RandomTest tree) {
	setCopy(new RandomTest(tree.probability));
    }

    public void visit(AndExpression tree) {
	tree.left.visit(this);
	Expression left = (Expression) getCopy();

	tree.right.visit(this);
	Expression right = (Expression) getCopy();

	setCopy(new AndExpression(left, right));
    }

    public void visit(OrExpression tree) {
	tree.left.visit(this);
	Expression left = (Expression) getCopy();

	tree.right.visit(this);
	Expression right = (Expression) getCopy();

	setCopy(new OrExpression(left, right));
    }
}
