/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.ast;

import java.util.ArrayList;


/**
 * Super class for all tree AST nodes of the Ant Control Language.
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public abstract class Tree {
    public abstract void visit(Visitor v);

    public Tree parent;

    /** 
     * Finds the enclosing {@link Block} of a {@link Tree tree}. For
     * most statements, this is <code>tree.parent</code>, if the
     * statement is not labeled.
     */
    public Block enclosingBlock() {
	Tree result = this;

	while (
	    (result != null) && 
	    (! ((result instanceof Block) || 
		(result instanceof Proc)))
	) {
	    result = result.parent;
	}

	if (result instanceof Proc) {
	    result = ((Proc) result).body;
	}

	return (Block) result;
    }

    public boolean isChildOf(Tree father) {
	for (Tree tree = this; tree != null; tree = tree.parent) {
	    if (tree == father) return true;
	}

	return false;
    }

    public static Tree newNopStatement() {
	return new Block(new Tree[] {});
    }

    /**
     * Tree class for representing labels <tt>name: statement;</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Label extends Tree {
	public String name;
	public Tree stmt;

	public Label(String name, Tree stmt) {
	    this.name = name;
	    this.stmt = stmt;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>var name;</tt>
     * statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Var extends Tree {
	public String name;

	public Var(String name) {
	    this.name = name;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>set var|unset var;</tt>
     * statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Set extends Tree {
	public String name;
	public boolean value;

	public Var definition;
	public String usageName;

	public Set(String name, boolean value) {
	    this.name = name;
	    this.value = value;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>var = expr;</tt>
     * statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Assign extends Tree {
	public String name;
	public Expression expr;

	public Var definition;

	public Assign(String name, Expression expr) {
	    this.name = name;
	    this.expr = expr;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <code>{...}</code> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Block extends Tree {
	public Tree[] stmts;

	private ArrayList localVars = new ArrayList();

	public Block(Tree[] stmts) {
	    this.stmts = stmts;
	}

	public void addLocalVariable(Var var) {
	    if (! localVars.contains(var)) {
		localVars.add(var);
	    }
	}

	public int getLocalVariableIndex(Var var) {
	    return localVars.indexOf(var);
	}

	public int getLocalVariableCnt() {
	    return localVars.size();
	}

	public Var getLocalVariable(int index) {
	    return (Var) localVars.get(index);
	}

	public void clearLocalVariables() {
	    localVars.clear();
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>if (expr) {...} else {};</tt>
     * statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class If extends Tree {
	public Expression expr;
	public Tree thenStmt;
	public Tree elseStmt;

	public If(Expression expr, Tree thenStmt, Tree elseStmt) {
	    this.expr = expr;
	    this.thenStmt = thenStmt;
	    this.elseStmt = elseStmt;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>goto label;</tt>
     * statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Goto extends Tree {
	public String name;
	public Label target;

	public Goto(String name) {
	    this.name = name;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>proc(var param, ...) {};</tt>
     * statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Proc extends Tree {
	public String name;
	public Var[] params;
	public Tree body;

	public Proc(String name, Var[] params, Tree body) {
	    this.name = name;
	    this.params = params;
	    this.body = body;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <code>foo(arg, ...)</code> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Call extends Tree {
	public String name;
	public Expression[] args;

	public Proc target;

	public Call(String name, Expression[] args) {
	    this.name = name;
	    this.args = args;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>mark n</tt> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class MarkCommand extends Tree {
	public int marker;

	public MarkCommand(int marker) {
	    this.marker = marker;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>unmark n</tt> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class UnmarkCommand extends Tree {
	public int marker;

	public UnmarkCommand(int marker) {
	    this.marker = marker;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>turn Left|Right</tt> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class TurnCommand extends Tree {
	public TurnDirection direction;

	public TurnCommand(TurnDirection direction) {
	    this.direction = direction;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>move</tt> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class MoveCommand extends Tree {
	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>pick</tt> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class PickUpCommand extends Tree {
	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing <tt>drop</tt> statements.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class DropCommand extends Tree {
	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing expressions.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public abstract static class Expression extends Tree {
    }

    /**
     * Tree node representing binary expressions.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public abstract static class BinaryExpression extends Expression {
	public Expression left;
	public Expression right;

	public BinaryExpression(Expression left, Expression right) {
	    this.left = left;
	    this.right = right;
	}
    }

    /**
     * Tree node representing the construct <tt>expression &&
     * expression</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class AndExpression extends BinaryExpression {
	public AndExpression(Expression left, Expression right) {
	    super(left, right);
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <tt>expression ||
     * expression</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class OrExpression extends BinaryExpression {
	public OrExpression(Expression left, Expression right) {
	    super(left, right);
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <code>! expression</code>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class NotExpression extends Expression {
	public Expression sub;

	public NotExpression(Expression sub) {
	    this.sub = sub;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <tt>Home[Ahead]</tt> and
     * so on.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class ConditionTest extends Expression {
	public SenseDirection direction;
	public Condition condition;

	public ConditionTest(SenseDirection direction, Condition condition) {
	    this.direction = direction;
	    this.condition = condition;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <tt>var</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Test extends Expression {
	public String name;
	public Var definition;

	public Test(String name) {
	    this.name = name;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <tt>true|false</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class Const extends Expression {
	public boolean value;

	public Const(boolean value) {
	    this.value = value;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <tt>HitWall</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class WallTest extends Expression {
	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <tt>FoundFood</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class FoodTest extends Expression {
	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Tree node representing the construct <tt>random(n)</tt>.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public static class RandomTest extends Expression {
	public int probability;

	public RandomTest(int probability) {
	    this.probability = probability;
	}

	public void visit(Visitor v) {
	    v.visit(this);
	}
    }

    /**
     * Visitor interface that is implemented by all {@link Tree}
     * nodes.
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public interface Visitor {
	public void visit(Label tree);
	public void visit(Block tree);
	public void visit(If tree);
	public void visit(Goto tree);
	public void visit(Proc tree);
	public void visit(Call tree);
	public void visit(MarkCommand tree);
	public void visit(UnmarkCommand tree);
	public void visit(TurnCommand tree);
	public void visit(MoveCommand tree);
	public void visit(PickUpCommand tree);
	public void visit(DropCommand tree);

	public void visit(Var tree);
	public void visit(Set tree);
	public void visit(Assign tree);
	public void visit(Test tree);
	public void visit(Const tree);

	public void visit(AndExpression tree);
	public void visit(OrExpression tree);
	public void visit(NotExpression tree);
	public void visit(ConditionTest tree);
	public void visit(WallTest tree);
	public void visit(FoodTest tree);
	public void visit(RandomTest tree);
    }
}
