/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.ast;
import ant.compiler.ast.Tree.*;
import ant.compiler.util.Printer;

import java.io.PrintWriter;

/**
 * Utility class for pretty printing AST representations of the Ant
 * Control Language.
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class TreePrinter extends Printer implements Tree.Visitor {
    private static final int INDENT = 4;

    PrintWriter out;
    int indentPos = 0;
    int pos = 0;
    final boolean debug;

    boolean newlineRequired = false;

    public TreePrinter(PrintWriter out) {
	this(out, false);
    }

    public TreePrinter(PrintWriter out, boolean debug) {
	this.out = out;	
	this.debug = debug;
    }

    public void visit(Label tree) {
	if (debug) println("// name=" + tree);
	println(tree.name + ":");
	tree.stmt.visit(this);
    }

    public void visit(Block tree) {
	println("{");
	indentPos += INDENT;
	if (debug) {
	    // DEBUG
	    if (tree.getLocalVariableCnt() > 0) {
		print("// local variables: ");
		for (int n = 0; n < tree.getLocalVariableCnt(); n++) {
		    Var var = tree.getLocalVariable(n);
		    if (n > 0) print(", ");
		    print(var.name);
		}
		println();
	    }

	    println("// name: " + tree);
	    if (tree.parent != null) {
		println("// parent: " + tree.parent);
	    }
	}

	for (int n = 0; n < tree.stmts.length; n++) {
	    tree.stmts[n].visit(this);
	}
	indentPos -= INDENT;
	print("}");
	newlineRequired = true;
    }

    public void visit(If tree) {
	print("if (");
	tree.expr.visit(this);
	print(") ");
	tree.thenStmt.visit(this);
	if (tree.elseStmt != null) {
	    if (newlineRequired) {
		newlineRequired = false;
		print(" ");
	    }

	    print("else ");
	    tree.elseStmt.visit(this);
	}
    }

    public void visit(Goto tree) {
	if (debug) println("// target=" + tree.target);
	println("goto " + tree.name + ";");
    }

    public void visit(Proc tree) {
	print("proc " + tree.name + "(");
	for (int n = 0; n < tree.params.length; n++) {
	    if (n > 0) print(", ");
	    tree.params[n].visit(this);
	    newlineRequired = false;
	}
	print(") ");
	tree.body.visit(this);
    }

    public void visit(Call tree) {
	print(tree.name + " (");
	for (int n = 0; n < tree.args.length; n++) {
	    if (n > 0) print(", ");
	    tree.args[n].visit(this);
	}
	println(");");
    }

    public void visit(MarkCommand tree) {
	println("mark " + tree.marker + ";");
    }

    public void visit(UnmarkCommand tree) {
	println("unmark " + tree.marker + ";");
    }

    public void visit(TurnCommand tree) {
	println("turn " + getLabel(tree.direction) + ";");
    }

    public void visit(MoveCommand tree) {
	println("move" + ";");
    }

    public void visit(PickUpCommand tree) {
	println("pick" + ";");
    }

    public void visit(DropCommand tree) {
	println("drop" + ";");
    }

    public void visit(Var tree) {
	print("var " + tree.name + (debug ? ID(tree) : ""));

	newlineRequired = true;
    }

    public void visit(Set tree) {
	printDef(tree.definition);
	println(tree.name + (debug ? ID(tree.definition) : "")  + " = " + tree.value + ";");
    }

    public void visit(Assign tree) {
	printDef(tree.definition);
	print(tree.name + (debug ? ID(tree.definition) : "") + " = ");
	tree.expr.visit(this);
	println(";");
    }

    public void visit(Test tree) {
	printDef(tree.definition);
	print(tree.name + (debug ? ID(tree.definition) : ""));
    }

    public void visit(Const tree) {
	print(Boolean.toString(tree.value));
    }

    public void visit(AndExpression tree) {
	print("(");
	tree.left.visit(this);
	print(") && (");
	tree.right.visit(this);
	print(")");
    }

    public void visit(OrExpression tree) {
	print("(");
	tree.left.visit(this);
	print(") || (");
	tree.right.visit(this);
	print(")");
    }

    public void visit(NotExpression tree) {
	print("!(");
	tree.sub.visit(this);
	print(")");
    }

    public void visit(ConditionTest tree) {
	print(getLabel(tree.condition) + 
	    (hasMarker(tree.condition) ? 
  	        Integer.toString(getMarker(tree.condition)) :
		"") +
	    "[" + getLabel(tree.direction) + "]");
    }

    public void visit(WallTest tree) {
	print("HitWall");
    }

    public void visit(FoodTest tree) {
	print("FoundFood");
    }

    public void visit(RandomTest tree) {
	print("random(1/" + tree.probability + ")");
    }

    private void indent() {
	if (pos == 0) {
	    for (int n = 0; n < indentPos; n++) {
		out.print(" ");
	    }
	    pos = indentPos;
	}
    }

    private void newline() {
	if (newlineRequired) {
	    out.println(";");
	    pos = 0;
	    newlineRequired = false;
	}
    }

    private void println() {
	newline();
	out.println();
	pos = 0;
    }

    private void println(String s) {
	newline();
	indent();
	out.println(s);
	pos = 0;
    }

    private void print(String s) {
	newline();
	indent();
	out.print(s);
	pos += s.length();
    }

    private String ID(Tree tree) {
	return "/*0x" + Integer.toHexString(System.identityHashCode(tree)) + "*/";
    }

    private void printDef(Tree definition) {
	print("/*def=" + definition + "*/");
    }

    public static void dumpTree(Tree tree, boolean debug) {
	PrintWriter out = new PrintWriter(System.out);
	if (debug) out.println("--- begin program dump ---");
 	tree.visit(new TreePrinter(out, debug));
	if (debug) out.println();
	if (debug) out.println("--- end program dump ---");
	if (debug) out.println();
	else out.println();
	out.flush();
    }
}
