/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.gen;

import ant.compiler.ast.Tree;
import ant.compiler.ast.Tree.*;
import ant.compiler.ast.DFSVisitor;

import ant.compiler.asm.Asm;
import ant.compiler.asm.AsmSymbols;
import ant.compiler.asm.Asm.*;

import java.util.HashMap;
import java.util.ArrayList;


/**
 * Generate {@link Asm} statements from an ACP abstract syntax {@link
 * Tree}.
 *
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class CodeGenerator extends DFSVisitor {
    Asm next;
    Asm result;

    AsmSymbols symbols;

    ExpressionEvaluator evaluator = new ExpressionEvaluator();

    public CodeGenerator(AsmSymbols symbols) {
	this.symbols = symbols;
    }

    public void reset() {
	next = new Asm.Nop();
    }

    public Asm getResult() {
	return result;
    }

    public void visit(Label tree) {
	super.visit(tree);
	symbols.registerLabel(tree.name, result);
    }

    public void visit(Block tree) {
	for (int n = tree.stmts.length - 1; n >= 0; n--) {
	    tree.stmts[n].visit(this);
	    next = result;
	}
	result = next;
    }

    public void visit(If tree) {
	Asm next = this.next;
	tree.thenStmt.visit(this);
	Asm thenOp = result;

	this.next = next;
	tree.elseStmt.visit(this);
	Asm elseOp = result;

	evaluator.reset();
	evaluator.setThen(thenOp);
	evaluator.setElse(elseOp);

	tree.expr.visit(evaluator);
	result = evaluator.getResult();
    }

    public void visit(Goto tree) {
	result = new Asm.Nop();
	symbols.registerJump(result, tree.name);
    }

    public void visit(MarkCommand tree) {
	result = new Asm.Mark(tree.marker, next);
    }

    public void visit(UnmarkCommand tree) {
	result = new Asm.Unmark(tree.marker, next);
    }

    public void visit(TurnCommand tree) {
	result = new Asm.Turn(tree.direction, next);
    }

    public void visit(MoveCommand tree) {
	result = new Asm.Move(next);
    }

    public void visit(PickUpCommand tree) {
	result = new Asm.PickUp(next);
    }

    public void visit(DropCommand tree) {
	result = new Asm.Drop(next);
    }

    public void visit(Tree tree) {
	throw new AssertionError(
	    "code generation impossible for " + tree.getClass().getName());
    }
}
