/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.trafo;

import ant.compiler.ast.Tree;
import ant.compiler.ast.Tree.*;
import ant.compiler.ast.Transformer;
import ant.compiler.util.ErrorLog;
import ant.compiler.util.Name;

import java.util.ArrayList;


/**
 * Visitor that computes the transformation for procedure
 * statements in the Ant Control Language.
 * 
 * <pre>proc name(pram, ...) {statements};</pre>
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class ProcTransformer extends Transformer {
    ErrorLog log;
    Tree root;

    SemanticAnalyzer semantic;

    int version = 0;
    Call currentCall;
    Proc currentProc;

    public ProcTransformer(ErrorLog log, Tree root) {
	this.semantic = new SemanticAnalyzer(log);
	this.log = log;
	this.root = root;
    }

    protected void afterUpdate(Tree before, Tree after) {
	if (before == root) root = after;
	root.visit(semantic);

	// Transform the transformed tree... :-))
	after.visit(this);
    }

    NameTransformer versionizer = new NameTransformer() {
	    protected String transformName(Label tree) {
		return Name.getVersionedCall(tree.name, version);
	    }

	    protected String transformName(Goto tree) {
		// bug workaround
		if (tree.target == null) return tree.name; 

		if (tree.target.isChildOf(currentProc)) {
		    return Name.getVersionedCall(tree.name, version);
		} else {
		    return tree.name; 
		}
	    }

	    protected String transformName(Var tree) {
		return Name.getVersionedCall(tree.name, version);
	    }

	    protected String transformName(Set tree) {
		// bug workaround
		if (tree.definition == null) return tree.name; 

		return versionedVarName(tree.name, tree, tree.definition);
	    }

	    protected String transformName(Assign tree) {
		// bug workaround
		if (tree.definition == null) return tree.name; 

		return versionedVarName(tree.name, tree, tree.definition);
	    }

	    protected String transformName(Test tree) {
		// bug workaround
		if (tree.definition == null) return tree.name; 

		return versionedVarName(tree.name, tree, tree.definition);
	    }

	    private String versionedVarName(String name, Tree tree, Var definition) {
		if (definition.isChildOf(currentProc)) {
		    return Name.getVersionedCall(name, version);
		}

		else if (currentCall.isChildOf(definition.enclosingBlock())) {
		    return name;
		}

		else {
		    log.error(
			"variable " + name + 
			" not visible when calling procedure " + currentProc.name);
		    return name;
		}
	    }

	    public void visit(Proc tree) {
		if (tree == currentProc) {
		    super.visit(tree);
		} else {
		    setCopy(
			Tree.newNopStatement());
		}
	    }

	};

    public void visit(Proc tree) {
	setResult(
	    Tree.newNopStatement());
    }

    public void visit(Call tree) {
	Proc copy;
	{
	    currentCall = tree;
	    currentProc = tree.target;

	    currentProc.visit(versionizer);
	    copy = (Proc) versionizer.getCopy();
	    version++;

	    currentCall = null;
	    currentProc = null;
	}

	ArrayList stmts = new ArrayList();
	for (int n = 0; n < copy.params.length; n++) {
	    stmts.add(copy.params[n]);
	}

	for (int n = 0; n < tree.args.length; n++) {
	    stmts.add(new Assign(copy.params[n].name, tree.args[n]));
	}

	stmts.add(copy.body);

	setResult(
	    new Block((Tree[]) stmts.toArray(new Tree[stmts.size()])));
    }
}
