/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.trafo;

import ant.compiler.ast.Tree;
import ant.compiler.ast.Tree.*;
import ant.compiler.ast.AbstractVisitor;
import ant.compiler.ast.DFSVisitor;
import ant.compiler.util.ErrorLog;
import ant.compiler.util.Name;

import java.util.HashMap;


/**
 * Visitor that performs the semantic analysis in Ant Control
 * Programs.
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class SemanticAnalyzer extends AbstractVisitor {
    ErrorLog log;

    HashMap variableDefinitions = new HashMap();
    HashMap labelDefinitions = new HashMap();
    HashMap procDefinitions = new HashMap();

    ParentResolver linker = new ParentResolver();
    DefinitionFinder finder = new DefinitionFinder();
    UsageResolver resolver = new UsageResolver();

    public SemanticAnalyzer(ErrorLog log) {
	this.log = log;
    }

    /**
     * Visitor that links tree nodes to their parents. 
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public class ParentResolver extends DFSVisitor {
	Tree parent;

	protected void beforeDescend(Tree tree) {
	    parent = tree;
	}

	protected void visitTree(Tree tree) {
	    tree.parent = parent;

	    // Note: Clear the parent after setting the parent of
	    // the current tree. Otherwise, the ParentResolver can
	    // not be reused and will write an endless
	    // .parent.parent... loop into the root element during
	    // the next run.
	    parent = null;
	}
    }

    /**
     * Visitor that finds definitions in the symbol table. 
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public class DefinitionFinder extends DFSVisitor {
	public void visit(Label tree) {
	    // System.err.println("DEBUG: label definition found: " + tree.name);
	    if (labelDefinitions.put(tree.name, tree) != null) {
		log.error("duplicate label: " + tree.name);
	    }
	    super.visit(tree);
	}

	public void visit(Var tree) {
	    // System.err.println("DEBUG: variable definition found: " + tree.name);
	    if (variableDefinitions.put(tree.name, tree) != null) {
		log.error("duplicate variable: " + tree.name);
	    }

	    Block enclosing = tree.enclosingBlock();

	    // 		System.err.println("DEBUG: vardef: ");
	    // 		TreePrinter.dumpTree(tree, true);
	    // 		System.err.println("DEBUG: enclosing: ");
	    // 		TreePrinter.dumpTree(enclosing, true);

	    enclosing.addLocalVariable(tree);
	    super.visit(tree);
	}

	public void visit(Proc tree) {
	    // System.err.println("DEBUG: procedure definition found: " + tree.name);
	    if (procDefinitions.put(tree.name, tree) != null) {
		log.error("duplicate procedure: " + tree.name);
	    }

	    super.visit(tree);
	}
    }

    /**
     * Visitor that enters definitions into the symbol table. 
     * 
     * <p>License: GPL</p>
     *
     * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
     */
    public class UsageResolver extends DFSVisitor {
	public void visit(Goto tree) {
	    if (tree.target == null) {
		tree.target = (Label) labelDefinitions.get(tree.name);
		if (tree.target == null) {
		    log.error("unresolved label: " + tree.name);
		}
	    }
	    super.visit(tree);
	}

	public void visit(Call tree) {
	    // Make sure, the call is not yet resolved, because
	    // during procedure transformation, multiple semantic
	    // analyses are performed, to attribute the
	    // substituted call statements.
	    if (tree.target == null) {
		tree.target = (Proc) procDefinitions.get(tree.name);
		if (tree.target == null) {
		    log.error("unresolved procedure: " + tree.name);
		}
	    }
	    super.visit(tree);
	}

	public void visit(Set tree) {
	    if (tree.definition == null) {
		tree.usageName = Name.nextUniqueUsage();
		tree.definition = (Var) variableDefinitions.get(tree.name);
		if (tree.definition == null) {
		    log.error("unresolved variable name in set: " + tree.name);
		}
	    }
	    super.visit(tree);
	}

	public void visit(Assign tree) {
	    if (tree.definition == null) {
		tree.definition = (Var) variableDefinitions.get(tree.name);
		if (tree.definition == null) {
		    log.error("unresolved variable name in assign: " + tree.name);
		}
	    }
	    super.visit(tree);
	}

	public void visit(Test tree) {
	    if (tree.definition == null) {
		tree.definition = (Var) variableDefinitions.get(tree.name);
		if (tree.definition == null) {
		    log.error("unresolved variable name in test: " + tree.name);
		}
	    }
	    super.visit(tree);
	}
    }
	
    protected void visitTree(Tree tree) {
	variableDefinitions.clear();
	labelDefinitions.clear();
	procDefinitions.clear();

	tree.visit(linker);
	tree.visit(finder);
	tree.visit(resolver);
    }
}
