/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.trafo;

import ant.compiler.ast.Tree;
import ant.compiler.ast.Tree.*;
import ant.compiler.ast.DFSVisitor;
import ant.compiler.util.Name;
import ant.compiler.util.ErrorLog;

import java.util.ArrayList;


/**
 * Visitor that computes the transformation for local variables in
 * the Ant Control Language.
 * 
 * <pre>var name;</pre>
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class VarTransformer extends DFSVisitor {
    Tree root;
    SemanticAnalyzer semantic;
    VersionCreator versionizer = new VersionCreator();

    public VarTransformer(ErrorLog log, Tree root) {
	this.semantic = new SemanticAnalyzer(log);
	this.root = root;
    }

    public void visit(Block tree) {
	// Transform inner blocks first. Oherwise, we need to keep
	// track of corresopnding blocks representing different
	// states of the same original variable.
	super.visit(tree);

	int varCnt = tree.getLocalVariableCnt();
	if (varCnt > 0) {
	    ArrayList newStmts = new ArrayList();

	    String endOfBlockName = Name.nextUnique();
		
	    int versionCnt = 1 << varCnt;

	    versionizer.setOriginal(tree);
	    for (int version = 0; version < versionCnt; version++) {
		versionizer.setVersion(version);
		tree.visit(versionizer);

		newStmts.add(versionizer.getCopy());
		newStmts.add(new Goto(endOfBlockName));
	    }
	    newStmts.add(new Label(endOfBlockName, Tree.newNopStatement()));

	    tree.stmts = (Tree[]) newStmts.toArray(new Tree[newStmts.size()]);
	    tree.clearLocalVariables();

	    // Note: There may exist jumps into the currently
	    // transformed block. It would be possible to fix that
	    // by redirecting them to the block with version zero,
	    // or even call such a block with predefined variable
	    // settings, but we delay that feature. So jumps into
	    // a block with local variables will result in
	    // unresolved labels!

	    // Reattribute the tree
	    root.visit(semantic);

	    // TreePrinter.dumpTree(root, true);
	}
    }
}
