/*
 * This file is part of the Ant Control Language compiler developed by
 * the team "KaNT" during the ICFP 2004 programming contest.
 *
 * ----------------------------------------------------------------------
 * 
 * Copyright (C) 2004  the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke
 * Thomas Moschny
 * Bernhard Haumacher
 * Bjoern-Oliver Hartmann
 *
 * ----------------------------------------------------------------------
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 * MA 02111-1307, USA.
 *
 */

package ant.compiler.trafo;

import ant.compiler.ast.Tree;
import ant.compiler.ast.Tree.*;
import ant.compiler.util.Name;


/**
 * Visitor class that creates a versiond copy of an Ant Control
 * Program fragment.
 * 
 * <p>License: GPL</p>
 *
 * @author <a href="http://www.haumacher.de/>">Bernhard Haumacher</a>
 */
public class VersionCreator extends NameTransformer {
    Block original;
    int version;

    public void setOriginal(Block original) {
	this.original = original;
    }

    public void setVersion(int version) {
	this.version = version;
    }

    protected String transformName(Label tree) {
	return Name.getVersioned(tree.name, version);
    }

    protected String transformName(Goto tree) {
	if (tree.target.isChildOf(original)) {
	    return Name.getVersioned(tree.name, version);
	} else {
	    return tree.name;
	}
    }

    protected String transformName(Var tree) {
	return Name.getVersioned(tree.name, version);
    }

    protected String transformName(Set tree) {
	if (tree.definition.isChildOf(original)) {
	    return Name.getVersioned(tree.name, version);
	} else {
	    return tree.name;
	}
    }

    protected String transformName(Test tree) {
	if (tree.definition.isChildOf(original)) {
	    return Name.getVersioned(tree.name, version);
	} else {
	    return tree.name;
	}
    }

    public void visit(Proc tree) {
	throw new AssertionError("procedure encountered while transforming variables");
    }

    public void visit(Call tree) {
	throw new AssertionError("call encountered while transforming variables");
    }

    public void visit(Assign tree) {
	throw new AssertionError("assign encountered while transforming variables");
    }

    public void visit(Var tree) {
	// Check, whether this variable is a direct child of the
	// currently transformed block.
	if (tree.enclosingBlock() == original) {
	    setCopy(Tree.newNopStatement());
	} else {
	    super.visit(tree);
	}
    }

    public void visit(Set tree) {
	// Check, whether the trees variable is currently
	// transformed.
	if (tree.definition.enclosingBlock() == original) {
	    int variableIndex = original.getLocalVariableIndex(tree.definition);

	    assert variableIndex >= 0 : "local variable";

	    int targetVersion = tree.value ? 
		(version |  (1 << variableIndex)) :
		(version & ~(1 << variableIndex));

	    String usageName = tree.usageName;
	    setCopy(
		new Block(new Tree[] {
		    new Goto(Name.getVersioned(usageName, targetVersion)),
		    new Label(Name.getVersioned(usageName, version), 
			Tree.newNopStatement())}));
	} else {
	    super.visit(tree);
	}
    }

    public void visit(Test tree) {
	// Check, whether the refered variable is currently
	// transformed.
	if (tree.definition.enclosingBlock() == original) {
	    int variableIndex = original.getLocalVariableIndex(tree.definition);

	    boolean constValue = 
		(version & (1 << variableIndex)) != 0;

	    setCopy(
		new Const(constValue));
	} else {
	    super.visit(tree);
	}
    }
}
