/*
 * ANT Simulation Programm
 *
 * ----------------------------------------------------------------------
 * 
 * (c) by the ICFP 2004 Team "KaNT"
 *
 * Jens Hauke <hauke@ipd.uka.de>
 * Thomas Moschny <moschny@ipd.uka.de>
 * Bernhard Haumacher <haui@haumacher.de>
 * Bjoern Hartmann <bjoern@derhartmann.de>
 *
 * ----------------------------------------------------------------------
 * 
 * This is free software (GPL).
 * 
 * $Rev: 1632 $
 * $Date: 2004-06-07 14:35:26 +0200 (Mon, 07 Jun 2004) $
 *
 * The main simulator engine.
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "sim.h"



ants_t ants;
world_t *world;

/* Just clean all markers. Must be called before init_antstat!!! */
static
void clean_world(void)
{
    int x, y;

    for (y = 0; y < MAX_WORLD_Y; y++) {
	for (x = 0; x < MAX_WORLD_X; x++) {
	    field_t *field = &world->field[y][x];
	    field->markers[ANT_RED] = 0;
	    field->markers[ANT_BLACK] = 0;
	    field->ant_id = NO_ANT;
	    field->food = field->original_food;
	}
    }
}

/* Call clean_world before !!! */
static
void init_antstat(antstat_t *antstat, antprog_t *ant,
		  unsigned int color,
		  unsigned int posy, unsigned int posx, unsigned int id)
{
    memset(antstat, 0, sizeof(*antstat)); /* for safty :-) */

    antstat->prog = ant;
    antstat->state = 0;
    antstat->pos.x = posx;
    antstat->pos.y = posy;
    antstat->dir = 0;
    antstat->color = color;
    antstat->resting = 0;
    antstat->has_food = 0;
    antstat->is_dead = 0;
    antstat->id = id;
    world->field[posy][posx].ant_id = id;
}

static
void init_ants(antprog_t *ant_red, antprog_t *ant_black)
{
    int x, y;
    ants.num_ants = 0;

    for (y = 0; y < MAX_WORLD_Y; y++) {
	for (x = 0; x < MAX_WORLD_X; x++) {
	    unsigned int id = ants.num_ants;
	    antstat_t *antstat = &ants.ant[id];
	    
	    if (world->field[y][x].home[ANT_RED]) {
		init_antstat(antstat, ant_red, ANT_RED, y, x, id);
		ants.num_ants++;
//		printf("red ant at %d, %d\n", x,y);
	    }
	    if (world->field[y][x].home[ANT_BLACK]) {
		init_antstat(antstat, ant_black, ANT_BLACK, y, x, id);
		ants.num_ants++;
//		printf("black ant at %d, %d\n", x,y);
	    }
	}
    }
}

const int dirx[2][6] =
{
    { +1, 0, -1, -1, -1, 0 }, /* even */
    { +1, +1, 0, -1,  0, +1} /* odd */
};

const int diry[6] =
{ 0, 1, 1 , 0, -1, -1};


static inline
pos_t adjacent_cell(pos_t pos, int dir)
{
    pos_t ret;

    ret.x = pos.x + dirx[pos.y % 2][dir]; 
    ret.y = pos.y + diry[dir];

    return ret;
}

static inline
int turn(int lr, int dir)
{
    return (dir + lr) % 6;
}


static inline
pos_t sensed_cell(pos_t pos, int dir, int sensedir)
{
    pos_t ret;
    
    switch (sensedir) {
    case SDIR_HERE:
	return pos;
    case SDIR_AHEAD:
	return adjacent_cell(pos, dir);
    case SDIR_LEFTAHEAD:
	return adjacent_cell(pos, turn(TURN_LEFT, dir));
    case SDIR_RIGHTAHEAD:
	return adjacent_cell(pos, turn(TURN_RIGHT, dir));
    default:
	printf("Internal error in %s : %d\n", __func__, __LINE__);
	exit(1);
    }
}

static inline
void set_marker_at(pos_t pos, int color, int marker)
{
    world->field[pos.y][pos.x].markers[color] |= (1 << marker);
}

static inline
void clear_marker_at(pos_t pos, int color, int marker)
{
    world->field[pos.y][pos.x].markers[color] &= ~(1 << marker);
}

static inline
int check_marker_at(pos_t pos, int color, int marker)
{
    return world->field[pos.y][pos.x].markers[color] & (1 << marker);
}

static inline
int check_any_marker_at(pos_t pos, int color)
{
    return world->field[pos.y][pos.x].markers[color];
}

static inline
int some_ant_at(pos_t pos)
{
    return world->field[pos.y][pos.x].ant_id != NO_ANT;
}

/* warning!!! no check inside */
static inline
antstat_t *ant_at(pos_t pos)
{
    int id = world->field[pos.y][pos.x].ant_id;
    return &ants.ant[id];
}

static inline
void set_ant_at(pos_t pos, antstat_t *ant)
{
    ant->pos = pos;
    world->field[pos.y][pos.x].ant_id = ant->id;
    ant->is_dead = 0;
}

static inline
void clear_ant_at(pos_t pos, antstat_t *ant)
{
    world->field[pos.y][pos.x].ant_id = NO_ANT;
    ant->is_dead = 1;
}

static inline
int anthill_at(pos_t pos, int color)
{
    return world->field[pos.y][pos.x].home[color];
}

static inline
int food_at(pos_t pos)
{
    return world->field[pos.y][pos.x].food;
}

static inline
void inc_food_at(pos_t pos)
{
    world->field[pos.y][pos.x].food++;
}

static inline
void dec_food_at(pos_t pos)
{
    world->field[pos.y][pos.x].food--;
}

/*
static inline
int ant_is_alive(antstat_t *ant)
{
    return !ant->is_dead;
}
*/

static inline
void kill_ant_at(pos_t pos)
{
    if (some_ant_at(pos))
	clear_ant_at(pos, ant_at(pos));
}

static inline
int rocky(pos_t pos)
{
    return world->field[pos.y][pos.x].rocky;
}


static inline
int other_color(int color)
{
    return 1 - color;
}

static inline
int cell_matches(pos_t pos, int cond, int i /* marker */, int color)
{
    if (rocky(pos)) {
	if (cond == COND_ROCK)
	    return 1;
	else
	    return 0;
    } else {
	switch (cond) {
	case COND_FRIEND:
	    return some_ant_at(pos) &&
		(ant_at(pos)->color == color);
	
	case COND_FOE:
	    return some_ant_at(pos) &&
		(ant_at(pos)->color != color);
	    
	case COND_FRIENDFOOD:
	    if (some_ant_at(pos)) {
		antstat_t *ant = ant_at(pos);

		return (ant->color == color) && ant->has_food;
	    } else {
		return 0;
	    }
	case COND_FOEFOOD:
	    if (some_ant_at(pos)) {
		antstat_t *ant = ant_at(pos);

		return (ant->color != color) && ant->has_food;
	    } else {
		return 0;
	    }
	    
	case COND_FOOD:
	    return food_at(pos) > 0;
	case COND_ROCK:
	    return 0;
	case COND_MARKER:
	    return check_marker_at(pos, color, i);
	case COND_FOEMARKER:
	    return check_any_marker_at(pos, other_color(color));
	case COND_HOME:
	    return anthill_at(pos, color);
	case COND_FOEHOME:
	    return anthill_at(pos, other_color(color));
	}
    }

    printf("Internal error in %s : %d\n", __func__, __LINE__);
    exit(1);
}

static inline
int next_s(int s)
{
    return s*22695477+1;
}

static unsigned int random_cur;

static
void randomseed(int seed)
{
    random_cur = next_s(next_s(next_s(seed)));
}

static
int randomint(int n)
{
    random_cur = next_s(random_cur);
    return ((random_cur / 65536) % 16384) % n;
}

static inline
int adjacent_ants(pos_t pos, int color)
{
    int d;
    int n = 0;

    for (d = 0; d < 6; d++) {
	pos_t cel = adjacent_cell(pos, d);
	if (some_ant_at(cel) &&
	    (ant_at(cel)->color == color)) {
	    n++;
	}
    }
    return n;
}

static inline
void check_for_surrounded_ants_at(pos_t pos)
{
    if (some_ant_at(pos)) {
	antstat_t *ant = ant_at(pos);
	if (adjacent_ants(pos, other_color(ant->color)) >= 5) {
	    if (ant->has_food) {
		inc_food_at(pos);
	    }
	    inc_food_at(pos);
	    inc_food_at(pos);
	    inc_food_at(pos);
	    kill_ant_at(pos);
	}
    }
}

static inline
void check_for_surrounded_ants(pos_t pos)
{
    int d;

    check_for_surrounded_ants_at(pos);

    for (d = 0; d < 6; d++) {
	check_for_surrounded_ants_at(adjacent_cell(pos, d));
    }
}

void halt_dummy(void)
{
    printf("%s", "");
}

int stepi;

static
void step(antstat_t *ant)
{
    antexpr_t *expr = &ant->prog->expr[ant->state];

//    if ((stepi == 3322) && (ant->id == 26)) {
    //halt_dummy();
    //}
    if (ant->is_dead) return;

    if (ant->resting) {
	ant->resting--;
	return;
    }
    
    switch (expr->token) {
    case TOKEN_ILLEGAL:
	printf("Ant of color %d exploded!!!!\n", ant->color);
	kill_ant_at(ant->pos);
	break;

    case TOKEN_SENSE:
	if (cell_matches(sensed_cell(ant->pos, ant->dir, expr->u.sense.sensedir),
			 expr->u.sense.cond,
			 expr->u.sense.i,
			 ant->color)) {
	    ant->state = expr->u.sense.st1;
	} else {
	    ant->state = expr->u.sense.st2;
	}
	break;
	
    case TOKEN_MARK:
	set_marker_at(ant->pos, ant->color, expr->u.mark.i);
	ant->state = expr->u.mark.st;
	break;
	
    case TOKEN_UNMARK:
	clear_marker_at(ant->pos, ant->color, expr->u.unmark.i);
	ant->state = expr->u.unmark.st;
	break;
	
    case TOKEN_PICKUP:
	if (ant->has_food || (food_at(ant->pos) == 0)) {
	    ant->state = expr->u.pickup.st2;
	} else {
	    dec_food_at(ant->pos);
	    ant->has_food = 1;
	    ant->state = expr->u.pickup.st1;
	}
	break;
	
    case TOKEN_DROP:
	if (ant->has_food) {
	    inc_food_at(ant->pos);
	    ant->has_food = 0;
	} 
	ant->state = expr->u.drop.st;
	break;
	
    case TOKEN_TURN:
	ant->dir = turn(expr->u.turn.lr, ant->dir);
	ant->state = expr->u.turn.st;
	break;
	
    case TOKEN_MOVE: {
	pos_t newp = adjacent_cell(ant->pos, ant->dir);

	if (rocky(newp) || some_ant_at(newp)) {
	    ant->state = expr->u.move.st2;
	} else {
	    clear_ant_at(ant->pos, ant);

	    set_ant_at(newp, ant);

	    ant->resting = 14;
	    ant->state = expr->u.move.st1;

	    check_for_surrounded_ants(newp);
	}
	break;
    }
    case TOKEN_FLIP:
	if (randomint(expr->u.flip.p) == 0) {
	    ant->state = expr->u.flip.st1;
	} else {
	    ant->state = expr->u.flip.st2;
	}
	break;
    default:
	printf("Internal error in %s : %d\n", __func__, __LINE__);
	exit(1);
    }
    
}

void dump_world()
{
    int x, y;

    for (y = 0; y < world->dimy; y++) {
	for (x = 0; x < world->dimx; x++) {
	    field_t *field = &world->field[y][x];
	    pos_t pos;
	    pos.x = x;
	    pos.y = y;
	    printf("cell (%d, %d): ", x, y);
	    if (field->rocky) printf("rock");
	    if (field->food) printf("%d food; ", field->food);
	    if (field->home[ANT_RED]) printf("red hill; ");
	    if (field->home[ANT_BLACK]) printf("black hill; ");

	    if (field->markers[ANT_RED]) {
		int b;
		printf("red marks: ");
		for (b = 0; b < 6; b++) {
		    if (field->markers[ANT_RED] & (1 << b))
			printf("%d", b);
		}
		printf("; ");
	    }
	    if (field->markers[ANT_BLACK]) {
		int b;
		printf("black marks: ");
		for (b = 0; b < 6; b++) {
		    if (field->markers[ANT_BLACK] & (1 << b))
			printf("%d", b);
		}
		printf("; ");
	    }

	    
	    if (some_ant_at(pos)) {
		antstat_t *ant = ant_at(pos);
		if (ant->color == ANT_RED) {
		    printf("red ");
		} else {
		    printf("black ");
		}
		printf("ant of id %d, dir %d, food %d, state %d, resting %d",
		       ant->id, ant->dir, ant->has_food, ant->state, ant->resting);
	    }
	    printf("\n");
	}
    }
}

void dump_world_debug()
{
    int antid;
    int x, y;

//    printf("# ant posx posy {0#red#|1#black#} {ID} {direction} {0|1#food#} {state} {0..13#resting#}\n");
    
    for (antid = 0; antid < ants.num_ants; antid++) {
	antstat_t *ant = &ants.ant[antid];

	if (!ant->is_dead) {
	    printf("ant %d %d %d %d %d %d %d %d\n",
		   ant->pos.x,
		   ant->pos.y,
		   ant->color,
		   ant->id,
		   ant->dir,
		   ant->has_food,
		   ant->state,
		   ant->resting);
	}
    }

    for (y = 0; y < world->dimy; y++) {
	for (x = 0; x < world->dimx; x++) {
	    field_t *field = &world->field[y][x];
	    if (field->food) {
		printf("food %d %d %d\n", x, y ,field->food);
	    }

	    if (field->markers[ANT_RED] || field->markers[ANT_BLACK]) {
		printf("marker %d %d %d %d\n",
		       x, y,
		       field->markers[ANT_RED],
		       field->markers[ANT_BLACK]);
	    }
	}
    }
}

int num_simsteps = 100000;

static inline
ratio_t calc_ratio(void)
{
    ratio_t res;
    int x, y;

    res.red = 0;
    res.black = 0;
    
    for (y = 0; y < world->dimy; y++) {
	for (x = 0; x < world->dimx; x++) {
	    field_t *field = &world->field[y][x];
	    if (field->home[ANT_RED]) {
		res.red += field->food;
	    }
	    if (field->home[ANT_BLACK]) {
		res.black += field->food;
	    }
	}
    }

    return res;
}

static inline
ratio_t count_ants(void)
{
    ratio_t res;
    int antid;
    res.red = 0;
    res.black = 0;

    for (antid = 0; antid < ants.num_ants; antid++) {
	antstat_t *ant = &ants.ant[antid];

	if (!ant->is_dead) {
	    if (ant->color == ANT_RED) {
		res.red += 1;
	    } else {
		res.black += 1;
	    }
	}
    }

    return res;
}

int num_dumpsteps = 1;

ratio_t simulate_alive(antprog_t *ant_red, antprog_t *ant_black, world_t *_world,
		       int dump, ratio_t *alive)
{

    int antid;
    
    world = _world;

    if (dump == DUMP_STD) {
	printf("random seed: 12345\n");
    }
    
    randomseed(12345);

    clean_world();

    init_ants(ant_red, ant_black);

    if (dump == DUMP_STD) {
	for (stepi = 0; stepi < num_simsteps; stepi++) {
	    printf("\nAfter round %d...\n", stepi);
	    dump_world();
	    for (antid = 0; antid < ants.num_ants; antid++) {
		antstat_t *ant = &ants.ant[antid];
		
		step(ant);
		
	    }
	}
    } else if (dump == DUMP_DEBUG) {
	for (stepi = 0; stepi < num_simsteps; stepi++) {
	    if ((stepi % num_dumpsteps) == 0) {
		printf("round %d\n", stepi);
		dump_world_debug();
	    }
	    for (antid = 0; antid < ants.num_ants; antid++) {
		antstat_t *ant = &ants.ant[antid];
		
		step(ant);
		
	    }
	}
    } else {
	/* fast loop, no dump */
	for (stepi = 0; stepi < num_simsteps; stepi++) {
	    for (antid = 0; antid < ants.num_ants; antid++) {
		antstat_t *ant = &ants.ant[antid];
		step(ant);
	    }
	}
    }

    if (alive) {
	*alive = count_ants();
    }
    return calc_ratio();
}

ratio_t simulate2_alive(antprog_t *ant_red, antprog_t *ant_black, world_t *_world, int dump,
			ratio_t *alive)
{
    ratio_t r1, r2;
    ratio_t alive2;
    
    r1 = simulate_alive(ant_red, ant_black, _world, dump, alive);
    /* exchange red and black */
    if (alive) {
	r2 = simulate_alive(ant_black, ant_red, _world, dump, &alive2);
	alive->red += alive2.black;
	alive->black += alive2.red;
    } else {
	r2 = simulate_alive(ant_black, ant_red, _world, dump, NULL);
    }

    
    r1.red += r2.black;
    r1.black += r2.red;
    
    return r1;
}

antprog_t *antprog_alloc(char *name)
{
    antprog_t *antp = calloc(1, sizeof(*antp));

    antp->name = name ? strdup(name) : NULL;

    return antp;
}

void antprog_setname(antprog_t *antp, char *name)
{
    if (antp->name) free(antp->name);

    antp->name = name ? strdup(name) : NULL;
}

void antprog_free(antprog_t *antp)
{
    if (antp->name) free(antp->name);
    free(antp);
}

world_t *world_alloc(void)
{
    int x, y;
    world_t *world = malloc(sizeof(*world));

    for (y = 0; y < MAX_WORLD_Y; y++) {
	for (x = 0; x < MAX_WORLD_X; x++) {
	    world->field[y][x].original_food = 0;
	    world->field[y][x].food = 0;
	    world->field[y][x].home[ANT_RED] = 0;
	    world->field[y][x].home[ANT_BLACK] = 0;
	    world->field[y][x].rocky = 0;
	}
    }
    world->dimx = 0;
    world->dimy = 0;

    return world;
}

void world_free(world_t *world)
{
    free(world);
}

/*
  Local Variables:
  c-basic-offset: 4
  End:
*/
